# -*- coding: utf-8 -*-
'''
Inkscape plugin for morphing maps to surveying stations a'la Therion

Useful if you need to update your map after closing a survey loop
or detecting a blunder.

Copyright (C) 2016 Mateusz Golicz, http://jaskinie.jaszczur.org
Distributed under the terms of the GNU General Public License v2
'''

import math
import sys
import inkex
import simpletransform
import logging
import simplepath
import copy
from operator import iadd
from speleo_kdtree import KDTree
from speleo import SpeleoEffect, SpeleoTransform


class SpeleoOutline(SpeleoEffect):
  def initOptions(self):
    self.OptionParser.add_option("--tolerance",
        action="store", type="float", 
        dest="tolerance", default=0.001)

  def addEndpoint(self, node, coords):
    '''
    Transform a point into global coordinates and add it to endpoint list
    '''
    sys.stderr.write("adding " + str(coords) + "\n")
    if self.transformCache <> None:
      tr = self.transformCache
    else:
      tr = self.transformCache = SpeleoTransform.getTotalTransform(node)
    self.endpoints.append(SpeleoTransform.transformCoords(coords, tr))
    
  def addEndpoints(self, node):
    # Parse path data
    path = simplepath.parsePath(node.get("d"))
    
    # Reset transform cache
    self.transformCache = None
    
    # Add endpoints of non-closed components
    lastM = None
    lastCmd = None
    
    for cmd in path:
      if cmd[0] == "M":
        if lastM: 
          self.addEndpoint(node, lastM[1])
          self.addEndpoint(node, lastCmd[1][-2:])
        lastM = cmd
      elif cmd[0] == "Z":
        lastM = None
      lastCmd = cmd
     
    if lastM:
      if lastM <> cmd:
        self.addEndpoint(node, lastM[1])
        self.addEndpoint(node, cmd[1][-2:])
        
  def findPaths(self, node, fn):
    for i in node:
      self.findPaths(i, fn)

    # If it's a path ...
    if node.tag == inkex.addNS('path','svg'):
      if node.get("id") <> "outline":
        fn(node)
  
  def processSubPath(self, path, xorVisible = False):
    # Cut path at breaking points
    if len(path) < 2: return path
    broken = False
    visible = True <> xorVisible
    closed = False
    prevP = False
    shoelace = 0
    ret = []
    for i in path:
      if len(i) == 3:
        broken = True
        # Breaking point
        if visible:
          # Breaking point at visible state, terminate drawing
          ret.append(i[0:2])
        else:
          # Breaking point at invisible state, move the cursor!
          ret.append(["M", i[1][-2:]])
        
        # Toggle visibility
        visible = not visible
      else:
        # Non-breaking point
        if visible:
          if (not broken) or i[0] != "Z":
            # Simply add this point
            ret.append(i)
          else:
            # Close the path
            closed = True
            ret.append(["L", path[0][1][-2:]])
      thisP = i[1][-2:]
      if len(thisP) == 2:
        if prevP:
          shoelace += (thisP[0] - prevP[0]) * (thisP[1] + prevP[1])
        prevP = thisP
    
    if (not xorVisible) and closed:
      thisP = path[0][1][-2:]
      shoelace += (thisP[0] - prevP[0]) * (thisP[1] + prevP[1])
      if shoelace < 0:
        return self.processSubPath(path, True)
        
    return ret
  
  def alterPath(self, node):
    # Parse path data
    path = simplepath.parsePath(node.get("d"))
    
    # Get total transform
    tr = SpeleoTransform.getTotalTransform(node)
    
    # Find path nodes near endpoints
    breakPoints = []
    for n, cmd in enumerate(path):
      # Not a coordinate command
      if len(cmd[1]) < 2: continue
      
      # Transform coordinates
      coords = SpeleoTransform.transformCoords(cmd[1][-2:], tr)
      
      # Find closest endpoint
      dist, index = self.kd.query(coords, 1);
      
      # Is this endpoint in the same place?
      if dist <= self.options.tolerance:
        breakPoints.append(n)
    
    # No intersections? No need to waste time!
    if len(breakPoints) == 0: return
    
    # OK, so this needs processing!

    # Mark the breakpoints
    for i in breakPoints:
      path[i].append(True)
      
    # Split into sub-paths
    subPaths = []
    prevM = 0
    for n, cmd in enumerate(path):
      if n > 0 and cmd[0] == "M":
        # New sub-path detected!
        subPaths.append(path[prevM:(n - 1)])
        prevM = n
    
    if len(subPaths):
      # Any subpaths in there?
      subPaths.append(path[prevM:])
      newPaths = []
      for i in subPaths:
        newPaths.append(self.processSubPath(i))
      path = reduce(iadd, newPaths)
    else:
      path = self.processSubPath(path)
    
    sys.stderr.write(str(path))
    node.set("d", simplepath.formatPath(path))
  
  def effect(self):
    # Configure self.logging
    if len(self.options.debug): logging.basicConfig(level = logging.DEBUG)
    
    self.endpoints = []
    self.findPaths(self.getRoot(), self.addEndpoints)

    self.kd = KDTree(self.endpoints)
    self.alterPath(self.getElementById("outline"))
    

    sys.stderr.write(str(self.endpoints) + "\n")
          
if __name__ == '__main__':
  e = SpeleoOutline()
  e.affect()

